@@ -1,14 +1,20 @@
 from functools import partial, reduce
 
+from django.core import checks
+from django.core.exceptions import ObjectDoesNotExist, FieldDoesNotExist
 from django.db import models
 from django.db.models import Max, Min, F
+from django.db.models.fields.related import ForeignKey
 from django.db.models.constants import LOOKUP_SEP
 from django.utils.module_loading import import_string
 from django.utils.translation import gettext_lazy as _
 
 
 def get_lookup_value(obj, field):
-    return reduce(lambda i, f: getattr(i, f), field.split(LOOKUP_SEP), obj)
+    try:
+        return reduce(lambda i, f: getattr(i, f), field.split(LOOKUP_SEP), obj)
+    except ObjectDoesNotExist:
+        return None
 
 
 class OrderedModelQuerySet(models.QuerySet):
@@ -19,21 +25,6 @@ class OrderedModelQuerySet(models.QuerySet):
         order_field_name = self._get_order_field_name()
         return LOOKUP_SEP.join([order_field_name, lookup])
 
-    def _get_order_with_respect_to(self):
-        model = self.model
-        order_with_respect_to = model.order_with_respect_to
-        if isinstance(order_with_respect_to, str):
-            order_with_respect_to = (order_with_respect_to,)
-        if order_with_respect_to is None:
-            raise AssertionError(
-                (
-                    'ordered model admin "{0}" has not specified "order_with_respect_to"; note that this '
-                    "should go in the model body, and is not to be confused with the Meta property of the same name, "
-                    "which is independent Django functionality"
-                ).format(model)
-            )
-        return order_with_respect_to
-
     def get_max_order(self):
         order_field_name = self._get_order_field_name()
         return self.aggregate(Max(order_field_name)).get(
@@ -90,41 +81,20 @@ class OrderedModelQuerySet(models.QuerySet):
 
     def bulk_create(self, objs, *args, **kwargs):
         order_field_name = self._get_order_field_name()
-        order_with_respect_to = self.model.order_with_respect_to
+        order_with_respect_to = self.model.get_order_with_respect_to()
         objs = list(objs)
-        if order_with_respect_to:
-            order_with_respect_to_mapping = {}
-            order_with_respect_to = self._get_order_with_respect_to()
-            for obj in objs:
-                key = tuple(
-                    get_lookup_value(obj, field) for field in order_with_respect_to
-                )
-                if key in order_with_respect_to_mapping:
-                    order_with_respect_to_mapping[key] += 1
-                else:
-                    order_with_respect_to_mapping[
-                        key
-                    ] = self.filter_by_order_with_respect_to(obj).get_next_order()
-                setattr(obj, order_field_name, order_with_respect_to_mapping[key])
-        else:
-            for order, obj in enumerate(objs, self.get_next_order()):
-                setattr(obj, order_field_name, order)
+        order_with_respect_to_mapping = {}
+        for obj in objs:
+            key = frozenset(obj._wrt_map().items())
+            if key in order_with_respect_to_mapping:
+                order_with_respect_to_mapping[key] += 1
+            else:
+                order_with_respect_to_mapping[key] = self.filter(
+                    **obj._wrt_map()
+                ).get_next_order()
+            setattr(obj, order_field_name, order_with_respect_to_mapping[key])
         return super().bulk_create(objs, *args, **kwargs)
 
-    def _get_order_with_respect_to_filter_kwargs(self, ref):
-        order_with_respect_to = self._get_order_with_respect_to()
-        _get_lookup_value = partial(get_lookup_value, ref)
-        return {field: _get_lookup_value(field) for field in order_with_respect_to}
-
-    _get_order_with_respect_to_filter_kwargs.queryset_only = False
-
-    def filter_by_order_with_respect_to(self, ref):
-        order_with_respect_to = self.model.order_with_respect_to
-        if order_with_respect_to:
-            filter_kwargs = self._get_order_with_respect_to_filter_kwargs(ref)
-            return self.filter(**filter_kwargs)
-        return self
-
 
 class OrderedModelManager(models.Manager.from_queryset(OrderedModelQuerySet)):
     pass
@@ -152,33 +122,77 @@ class OrderedModelBase(models.Model):
     class Meta:
         abstract = True
 
-    def _validate_ordering_reference(self, ref):
-        if self.order_with_respect_to is not None:
-            self_kwargs = (
-                self._meta.default_manager._get_order_with_respect_to_filter_kwargs(
-                    self
-                )
-            )
-            ref_kwargs = (
-                ref._meta.default_manager._get_order_with_respect_to_filter_kwargs(ref)
-            )
-            if self_kwargs != ref_kwargs:
-                raise ValueError(
-                    "{0!r} can only be swapped with instances of {1!r} with equal {2!s} fields.".format(
-                        self,
-                        self._meta.default_manager.model,
-                        " and ".join(["'{}'".format(o) for o in self_kwargs]),
-                    )
-                )
+    def __init__(self, *args, **kwargs):
+        super(OrderedModelBase, self).__init__(*args, **kwargs)
+        self._original_wrt_map = self._wrt_map()
 
-    def get_ordering_queryset(self, qs=None):
+    def _wrt_map(self):
+        d = {}
+        for order_wrt_name in self.get_order_with_respect_to():
+            # we know order_wrt_name is a ForeignKey, so use a cheaper _id lookup
+            field_path = order_wrt_name + "_id"
+            d[order_wrt_name] = get_lookup_value(self, field_path)
+        return d
+
+    def _get_related_objects(self):
+        # slow path, for use in the admin which requires the objects
+        # expected to generate extra queries
+        return [
+            get_lookup_value(self, name) for name in self.get_order_with_respect_to()
+        ]
+
+    @classmethod
+    def _on_ordered_model_delete(cls, sender=None, instance=None, **kwargs):
+        """
+        This signal handler makes sure that when an OrderedModelBase is deleted via
+        cascade database deletes, or queryset delete that the models keep order.
+        """
+
+        if getattr(instance, "_was_deleted_via_delete_method", False):
+            return
+
+        extra_update = kwargs.get("extra_update", None)
+
+        # Copy of upshuffle logic from OrderedModelBase.delete
+        qs = instance.get_ordering_queryset()
+        extra_update = {} if extra_update is None else extra_update
+        qs.above_instance(instance).decrease_order(**extra_update)
+
+        setattr(instance, "_was_deleted_via_delete_method", True)
+
+    @classmethod
+    def get_order_with_respect_to(cls):
+        if type(cls.order_with_respect_to) is tuple:
+            return cls.order_with_respect_to
+        elif type(cls.order_with_respect_to) is str:
+            return (cls.order_with_respect_to,)
+        elif cls.order_with_respect_to is None:
+            return tuple()
+        else:
+            raise ValueError("Invalid value for model.order_with_respect_to")
+
+    def _validate_ordering_reference(self, ref):
+        if self._wrt_map() != ref._wrt_map():
+            raise ValueError(
+                "{0!r} can only be swapped with instances of {1!r} with equal {2!s} fields.".format(
+                    self,
+                    self._meta.default_manager.model,
+                    " and ".join(
+                        ["'{}'".format(o) for o in self.get_order_with_respect_to()]
+                    ),
+                )
+            )
+
+    def get_ordering_queryset(self, qs=None, wrt=None):
         if qs is None:
             if self.order_class_path:
                 model = import_string(self.order_class_path)
                 qs = model._meta.default_manager.all()
             else:
                 qs = self._meta.default_manager.all()
-        return qs.filter_by_order_with_respect_to(self)
+        if wrt:
+            return qs.filter(**wrt)
+        return qs.filter(**self._wrt_map())
 
     def previous(self):
         """
@@ -194,12 +208,25 @@ class OrderedModelBase(models.Model):
 
     def save(self, *args, **kwargs):
         order_field_name = self.order_field_name
-        if getattr(self, order_field_name) is None:
+        wrt_changed = self._wrt_map() != self._original_wrt_map
+
+        if wrt_changed and getattr(self, order_field_name) is not None:
+            # do delete-like upshuffle using original_wrt values!
+            qs = self.get_ordering_queryset(wrt=self._original_wrt_map)
+            qs.above_instance(self).decrease_order()
+
+        if getattr(self, order_field_name) is None or wrt_changed:
             order = self.get_ordering_queryset().get_next_order()
             setattr(self, order_field_name, order)
         super().save(*args, **kwargs)
 
+        self._original_wrt_map = self._wrt_map()
+
     def delete(self, *args, extra_update=None, **kwargs):
+        # Flag re-ordering performed so that post_delete signal
+        # does not duplicate the re-ordering. See signals.py
+        self._was_deleted_via_delete_method = True
+
         qs = self.get_ordering_queryset()
         extra_update = {} if extra_update is None else extra_update
         qs.above_instance(self).decrease_order(**extra_update)
@@ -307,6 +334,91 @@ class OrderedModelBase(models.Model):
         o = self.get_ordering_queryset().get_max_order()
         self.to(o, extra_update=extra_update)
 
+    @classmethod
+    def check(cls, **kwargs):
+        errors = super().check(**kwargs)
+
+        ordering = getattr(cls._meta, "ordering", None)
+        if ordering is None or len(ordering) < 1:
+            errors.append(
+                checks.Error(
+                    "OrderedModelBase subclass needs Meta.ordering specified.",
+                    hint="If you have overwritten Meta, try inheriting with Meta(OrderedModel.Meta).",
+                    obj=str(cls.__qualname__),
+                    id="ordered_model.E001",
+                )
+            )
+        owrt = getattr(cls, "order_with_respect_to")
+        if not (type(owrt) is tuple or type(owrt) is str or owrt is None):
+            errors.append(
+                checks.Error(
+                    "OrderedModelBase subclass order_with_respect_to value invalid. Expected tuple, str or None.",
+                    obj=str(cls.__qualname__),
+                    id="ordered_model.E002",
+                )
+            )
+        if not issubclass(cls.objects.__class__, OrderedModelManager):
+            # Not using our Manager. This is an Error if the queryset is also wrong, or
+            # a Warning if our own QuerySet is returned.
+            if issubclass(cls.objects.none().__class__, OrderedModelQuerySet):
+                errors.append(
+                    checks.Warning(
+                        "OrderedModelBase subclass has a ModelManager that does not inherit from OrderedModelManager. This is not ideal but will work.",
+                        obj=str(cls.__qualname__),
+                        id="ordered_model.W003",
+                    )
+                )
+            else:
+                errors.append(
+                    checks.Error(
+                        "OrderedModelBase subclass has a ModelManager that does not inherit from OrderedModelManager.",
+                        obj=str(cls.__qualname__),
+                        id="ordered_model.E003",
+                    )
+                )
+        elif not issubclass(cls.objects.none().__class__, OrderedModelQuerySet):
+            errors.append(
+                checks.Error(
+                    "OrderedModelBase subclass ModelManager did not return a QuerySet inheriting from OrderedModelQuerySet.",
+                    obj=str(cls.__qualname__),
+                    id="ordered_model.E004",
+                )
+            )
+
+        # each field may be an FK, or recursively an FK ref to an FK
+        try:
+            for wrt_field in cls.get_order_with_respect_to():
+                mc = cls
+                for p in wrt_field.split(LOOKUP_SEP):
+                    try:
+                        f = mc._meta.get_field(p)
+                        if not isinstance(f, ForeignKey):
+                            errors.append(
+                                checks.Error(
+                                    "OrderedModel order_with_respect_to specifies field '{0}' (within '{1}') which is not a ForeignKey. This is unsupported.".format(
+                                        p, wrt_field
+                                    ),
+                                    obj=str(cls.__qualname__),
+                                    id="ordered_model.E005",
+                                )
+                            )
+                            break
+                        mc = f.remote_field.model
+                    except FieldDoesNotExist:
+                        errors.append(
+                            checks.Error(
+                                "OrderedModel order_with_respect_to specifies field '{0}' (within '{1}') which does not exist.".format(
+                                    p, wrt_field
+                                ),
+                                obj=str(cls.__qualname__),
+                                id="ordered_model.E006",
+                            )
+                        )
+        except ValueError:
+            # already handled by type checks for E002
+            pass
+        return errors
+
 
 class OrderedModel(OrderedModelBase):
     """
