@@ -3,14 +3,24 @@ from io import StringIO
 
 from django.contrib.auth.models import User
 from django.core.management import call_command
+from django.core import checks
+from django.db import models
+from django.db.models.signals import post_delete
+from django.dispatch import Signal
 from django.utils.timezone import now
 from django.urls import reverse
-from django.test import TestCase
+from django.test import TestCase, SimpleTestCase
+from django.test.utils import isolate_apps, override_system_checks
 from django import VERSION
 
+
 from rest_framework.test import APIRequestFactory, APITestCase
 from rest_framework import status
 from tests.drf import ItemViewSet, router
+from tests.utils import assertNumQueries
+
+from ordered_model.models import OrderedModel, OrderedModelManager, OrderedModelQuerySet
+
 
 from tests.models import (
     Answer,
@@ -23,11 +33,14 @@ from tests.models import (
     Pizza,
     Topping,
     PizzaToppingsThroughModel,
+    BaseQuestion,
     OpenQuestion,
     MultipleChoiceQuestion,
     ItemGroup,
     GroupedItem,
     TestUser,
+    CascadedParentModel,
+    CascadedOrderedModel,
 )
 
 
@@ -154,6 +167,9 @@ class OrderWithRespectToTests(TestCase):
         self.q1_a2 = q1.answers.create(user=u0)
         self.q2_a2 = q2.answers.create(user=u0)
 
+        with assertNumQueries(self, 1):
+            Answer.objects.get(id=self.q1_a1.id)
+
     def test_saved_order(self):
         self.assertSequenceEqual(
             Answer.objects.values_list("pk", "order"),
@@ -281,6 +297,39 @@ class OrderWithRespectToTests(TestCase):
         )
 
 
+class OrderWithRespectToReorderTests(TestCase):
+    def setUp(self):
+        q1 = Question.objects.create()
+        self.u0 = TestUser.objects.create()
+        self.u1 = TestUser.objects.create()
+        self.u0_a1 = q1.answers.create(user=self.u0)
+        self.u0_a2 = q1.answers.create(user=self.u0)
+        self.u0_a3 = q1.answers.create(user=self.u0)
+        self.u1_a1 = q1.answers.create(user=self.u1)
+        self.u1_a2 = q1.answers.create(user=self.u1)
+        self.u1_a3 = q1.answers.create(user=self.u1)
+
+        with assertNumQueries(self, 1):
+            Answer.objects.get(id=self.u0_a1.id)
+
+    def test_reorder_when_field_value_changed(self):
+        self.u0_a2.user = self.u1
+        with assertNumQueries(self, 3):
+            self.u0_a2.save()
+
+        self.assertSequenceEqual(
+            Answer.objects.values_list("pk", "order"),
+            [
+                (self.u0_a1.pk, 0),
+                (self.u0_a3.pk, 1),
+                (self.u1_a1.pk, 0),
+                (self.u1_a2.pk, 1),
+                (self.u1_a3.pk, 2),
+                (self.u0_a2.pk, 3),
+            ],
+        )
+
+
 class CustomPKTest(TestCase):
     def setUp(self):
         self.item1 = CustomItem.objects.create(pkid=str(uuid.uuid4()), name="1")
@@ -647,7 +696,9 @@ class OrderWithRespectToTestsManyToMany(TestCase):
         )  # tomatoe, mozarella, mushrooms, ham
         # Now put the toppings on the pizza
         self.p1_t1 = PizzaToppingsThroughModel(pizza=self.p1, topping=self.t1)
-        self.p1_t1.save()
+        with assertNumQueries(self, 2):
+            self.p1_t1.save()
+
         self.p1_t2 = PizzaToppingsThroughModel(pizza=self.p1, topping=self.t2)
         self.p1_t2.save()
         self.p1_t3 = PizzaToppingsThroughModel(pizza=self.p1, topping=self.t3)
@@ -658,7 +709,9 @@ class OrderWithRespectToTestsManyToMany(TestCase):
         self.p2_t2.save()
         self.p2_t3 = PizzaToppingsThroughModel(pizza=self.p2, topping=self.t4)
         self.p2_t3.save()
-        self.p2_t4 = PizzaToppingsThroughModel(pizza=self.p2, topping=self.t5)
+        self.p2_t4 = PizzaToppingsThroughModel()
+        self.p2_t4.pizza = self.p2
+        self.p2_t4.topping = self.t5
         self.p2_t4.save()
 
     def test_saved_order(self):
@@ -1073,6 +1126,45 @@ class ReorderModelTestCase(TestCase):
             "changing order of tests.GroupedItem (3) from 1 to 2\n", out.getvalue()
         )
 
+    def test_reorder_with_respect_to_tuple(self):
+        u1 = TestUser.objects.create()
+        u2 = TestUser.objects.create()
+        q1 = Question.objects.create()
+        q2 = Question.objects.create()
+
+        for q in (q1, q2):
+            for u in (u1, u2):
+                Answer.objects.create(user=u, question=q, order=0)
+                Answer.objects.create(user=u, question=q, order=0)
+
+        self.assertSequenceEqual(
+            Answer.objects.filter(user=u2, question=q1).values_list("order", flat=True),
+            [0, 0],
+        )
+
+        out = StringIO()
+        call_command("reorder_model", "tests.Answer", verbosity=1, stdout=out)
+
+        self.assertSequenceEqual(
+            Answer.objects.filter(user=u2, question=q1).values_list("order", flat=True),
+            [0, 1],
+        )
+
+        self.assertEqual(
+            (
+                "changing order of tests.Answer (2) from 0 to 1\n"
+                + "changing order of tests.Answer (4) from 0 to 1\n"
+                + "changing order of tests.Answer (6) from 0 to 1\n"
+                + "changing order of tests.Answer (8) from 0 to 1\n"
+            ),
+            out.getvalue(),
+        )
+
+        out = StringIO()
+        call_command("reorder_model", "tests.Answer", verbosity=1, stdout=out)
+
+        self.assertEqual("", out.getvalue())
+
     def test_reorder_with_custom_order_field(self):
         """
         Test that 'reorder_model' changes the order of OpenQuestions
@@ -1111,7 +1203,26 @@ class ReorderModelTestCase(TestCase):
         OpenQuestion.objects.create(answer="4", order=3)
 
         # bypass our OrderedModel delete logic to leave a hole in ordering
+        # remove signal handlers
+        # print(post_delete.receivers)
+        self.assertTrue(
+            post_delete.disconnect(
+                sender=OpenQuestion, dispatch_uid=OpenQuestion.__name__
+            )
+        )
+        self.assertTrue(
+            post_delete.disconnect(
+                sender=BaseQuestion, dispatch_uid=BaseQuestion.__name__
+            )
+        )
+
+        # delete on the  queryset fires post_delete, but does not call model.delete()
         OpenQuestion.objects.filter(answer="3").delete()
+        post_delete.connect(
+            OpenQuestion._on_ordered_model_delete,
+            sender=OpenQuestion,
+            dispatch_uid=OpenQuestion.__name__,
+        )
 
         self.assertEqual([0, 1, 3], [i.order for i in OpenQuestion.objects.all()])
         self.assertEqual(
@@ -1211,3 +1322,173 @@ class DRFTestCase(APITestCase):
         self.assertEqual(response.data, {"pkid": "b", "name": "2", "renamedOrder": 0})
         self.assertEqual(CustomItem.objects.get(pkid="b").order, 0)
         self.assertEqual(CustomItem.objects.get(pkid="a").order, 1)
+
+
+@isolate_apps("tests", attr_name="apps")
+@override_system_checks([checks.model_checks.check_all_models])
+class ChecksTest(SimpleTestCase):
+    def test_no_inherited_ordering(self):
+        class TestModel(OrderedModel):
+            class Meta:
+                verbose_name = "unordered"
+
+        self.maxDiff = None
+        self.assertEqual(
+            checks.run_checks(app_configs=self.apps.get_app_configs()),
+            [
+                checks.Error(
+                    msg="OrderedModelBase subclass needs Meta.ordering specified.",
+                    hint="If you have overwritten Meta, try inheriting with Meta(OrderedModel.Meta).",
+                    obj="ChecksTest.test_no_inherited_ordering.<locals>.TestModel",
+                    id="ordered_model.E001",
+                )
+            ],
+        )
+
+    def test_explicit_ordering(self):
+        class TestModel2(OrderedModel):
+            class Meta:
+                verbose_name = "unordered"
+                ordering = ["order"]
+
+        self.assertEqual(checks.run_checks(app_configs=self.apps.get_app_configs()), [])
+
+    def test_inherited_ordering(self):
+        class TestModel(OrderedModel):
+            class Meta(OrderedModel.Meta):
+                verbose_name = "unordered"
+
+        self.assertEqual(checks.run_checks(app_configs=self.apps.get_app_configs()), [])
+
+    def test_bad_owrt(self):
+        class TestModel(OrderedModel):
+            order_with_respect_to = 7
+
+        self.assertEqual(
+            checks.run_checks(app_configs=self.apps.get_app_configs()),
+            [
+                checks.Error(
+                    msg="OrderedModelBase subclass order_with_respect_to value invalid. Expected tuple, str or None.",
+                    obj="ChecksTest.test_bad_owrt.<locals>.TestModel",
+                    id="ordered_model.E002",
+                )
+            ],
+        )
+
+    def test_bad_manager(self):
+        class BadModelManager(models.Manager.from_queryset(models.QuerySet)):
+            pass
+
+        class TestModel(OrderedModel):
+            objects = BadModelManager()
+
+        self.assertEqual(
+            checks.run_checks(app_configs=self.apps.get_app_configs()),
+            [
+                checks.Error(
+                    msg="OrderedModelBase subclass has a ModelManager that does not inherit from OrderedModelManager.",
+                    obj="ChecksTest.test_bad_manager.<locals>.TestModel",
+                    id="ordered_model.E003",
+                )
+            ],
+        )
+
+    def test_builtin_manager_to_queryset(self):
+        class TestModel(OrderedModel):
+            objects = OrderedModelQuerySet.as_manager()
+
+        self.assertEqual(
+            checks.run_checks(app_configs=self.apps.get_app_configs()),
+            [
+                checks.Warning(
+                    msg="OrderedModelBase subclass has a ModelManager that does not inherit from OrderedModelManager. This is not ideal but will work.",
+                    obj="ChecksTest.test_builtin_manager_to_queryset.<locals>.TestModel",
+                    id="ordered_model.W003",
+                )
+            ],
+        )
+
+    def test_bad_queryset(self):
+        # I've swapped the inheritance order here so that the models.QuerySet is returned
+        class BadQSModelManager(
+            models.Manager.from_queryset(models.QuerySet), OrderedModelManager
+        ):
+            pass
+
+        class TestModel(OrderedModel):
+            objects = BadQSModelManager()
+
+        self.assertEqual(
+            checks.run_checks(app_configs=self.apps.get_app_configs()),
+            [
+                checks.Error(
+                    msg="OrderedModelBase subclass ModelManager did not return a QuerySet inheriting from OrderedModelQuerySet.",
+                    obj="ChecksTest.test_bad_queryset.<locals>.TestModel",
+                    id="ordered_model.E004",
+                )
+            ],
+        )
+
+    def test_owrt_not_foreign_key(self):
+        class TestModel(OrderedModel):
+            name = models.CharField(max_length=100)
+            order_with_respect_to = "name"
+
+        self.assertEqual(
+            checks.run_checks(app_configs=self.apps.get_app_configs()),
+            [
+                checks.Error(
+                    msg="OrderedModel order_with_respect_to specifies field 'name' (within 'name') which is not a ForeignKey. This is unsupported.",
+                    obj="ChecksTest.test_owrt_not_foreign_key.<locals>.TestModel",
+                    id="ordered_model.E005",
+                )
+            ],
+        )
+
+    def test_owrt_not_immediate_foreign_key(self):
+        class TestTargetModel(OrderedModel):
+            name = models.CharField(max_length=100)
+
+        class TestModel(OrderedModel):
+            target = models.ForeignKey(to=TestTargetModel, on_delete=models.CASCADE)
+            order_with_respect_to = "target__name"
+
+        self.assertEqual(
+            checks.run_checks(app_configs=self.apps.get_app_configs()),
+            [
+                checks.Error(
+                    msg="OrderedModel order_with_respect_to specifies field 'name' (within 'target__name') which is not a ForeignKey. This is unsupported.",
+                    obj="ChecksTest.test_owrt_not_immediate_foreign_key.<locals>.TestModel",
+                    id="ordered_model.E005",
+                )
+            ],
+        )
+
+
+class TestCascadedDelete(TestCase):
+    def test_that_model_when_deleted_by_cascade_still_maintains_ordering(self):
+        parent_for_order_0_child = CascadedParentModel.objects.create()
+        child_with_order_0 = CascadedOrderedModel.objects.create(
+            parent=parent_for_order_0_child
+        )
+
+        parent__for_order_1_child = CascadedParentModel.objects.create()
+        child_with_order_1 = CascadedOrderedModel.objects.create(
+            parent=parent__for_order_1_child
+        )
+
+        parent_for_order_2_child = CascadedParentModel.objects.create()
+        child_with_order_2 = CascadedOrderedModel.objects.create(
+            parent=parent_for_order_2_child
+        )
+
+        # Delete positition 1 parent, now there's a hole, which child_with_order_2 should take
+        parent__for_order_1_child.delete()
+
+        # Refresh children from db
+        child_with_order_0.refresh_from_db()
+        child_with_order_2.refresh_from_db()
+
+        # Assert the hole has been filled
+        self.assertEqual(child_with_order_0.order, 0)
+        self.assertEqual(child_with_order_2.order, 1)
